/** \file seti.c
 *  
 *  SETI@home plugin for GKrellM
 *
 *  Copyright (c) 2000 Xavier Serpaggi
 *
 *  Authors: Xavier Serpaggi < xavier dot serpaggi at free dot fr >
 *      and  Tilman Sauerbeck < tilman dot sauerbeck at web dot de > (see README-pcpu)
 *
 *  Version: 0.7.0b
 *
 *  This program is free software which I release under the GNU General Public
 *  License. You may redistribute and/or modify this program under the terms
 *  of that license as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* seti.c       - requires GKrellM 1.2.2 or later.
 *  If this file is compiled as a shared object file and installed in
 *  in your ~/.gkrellm2/plugins directory, then it will be loaded when
 *  gkrellm starts up.
 *     gcc -fPIC `pkg-config gtk+-2.0 --cflags`  -c seti.c
 *     gcc -shared -Wl -o seti.so seti.o
 *     cp seti.so ~/.gkrellm2/plugins
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <locale.h>

#include "seti.h"
#include "utils.h"


struct lconv		*locale_conv ;
struct stat	 	 seti_stat_buf ;
gint			 seti_paths_status ;
T_user_info		 user_info ;
T_client_info		 client_info ;
T_work_unit_info	 work_unit_info ;
gchar			 buf[256] ;
gdouble			 seconds[2] ;
gdouble			 total_time[2] ;
gboolean		 my_switch ;



static gchar		 tooltip_text[64] ;

static GtkWidget	*seti_path_entry ;
static GtkWidget	*seti_args_entry ;
static GtkWidget	*seti_client_start_entry ;
static GtkWidget	*seti_client_stop_entry ;
static GtkWidget	*seti_client_entry;
static GtkWidget	*pcpu_select_button[3];
static GtkTooltips	*tooltip ;

static GkrellmPiximage	*krell_image_seti ;
static GkrellmPiximage	*bg_meter_image_seti ;
static GkrellmPiximage	*krell_image_pcpu ;
static GkrellmPiximage	*bg_meter_image_pcpu ;

static GkrellmMonitor		*monitor ;
static GkrellmPanel		*panel;
static GkrellmPanel		*panel_pcpu ;
static gint		 style_id ;

static GkrellmKrell		*k;
static GkrellmKrell		*k_pcpu ;
static GkrellmKrell		*k_pcpu2 ;
static GkrellmDecal		*decal_text_pcpu ;
static gchar		 pcpu_text[10] ;

static gint		 pcpu_display ;
static gboolean		 krell_inserted ;

static gchar		*seti_status_report[] = {_("no workunit"), _("downloading"), _("processing"), _("done!")} ;
static gchar		*seti_info_text[] =
{
	_("<b>SETI@home progress meter:\n\n"),
	_("SETI@home progress meter shows the current computation progress of\n"),
	_("a SETI@home work unit (see http://setiathome.ssl.berkeley.edu for\n"),
	_("more information about SETI@home).\n\n"),
	_("Number of work units processed, current work unit completion percent\n"),
	_("and status, is available via a tooltip.\n\n"),
	_("Progress value is read every minute, as client write his results\n"),
	_("every minute.\n\n"),
	_("CPU Usage of the SETI@home process can be displayed in two modes,\n"),
	_("switch between them by middle clicking the SETI@home panel\n"),
	_("resp. use the "),
	_("<i>CPU Usage"),
	_(" tab.\n\n"),
	_("User can change paths to the SETI@home client and state.sah file\n"),
	_("using the "),
	_("<i>Configs"),
	_(" tab.\n\n"),
	_("SETI@home client can be started and stopped by pressing left mouse\n"),
	_("button over the SETI@home plugin monitor.\n"),
	_("In the "),
	_("<i>Client Control"),
	_(" tab, you can give names of programs or scripts\n"),
	_("to start and/or stop the client. They will be executed the way you have\n"),
	_("written them, without additional arguments.\n"),
	_("If you want to use internal mechanisms, just leave the \"internal\" value.\n"),
	_("This will start and stop the client as decribed in the setiathome\n"),
	_("distribution README file.\n"),
	_("Arguments passed to the client when started this way may be changed in\n"),
	_("the "),
	_("<i>Configs"),
	_(" tab.\n"),
	_("Note:\n"),
	_("  if you use setiathome through a proxy don't forget to add the\n"),
	_("  "),
	_("<i>-proxy hostname:port"),
	_(" option. Without it the client won't be able to send\n"),
	_("or receive data.\n\n"),
	_("Information about the user, as displayed on the SETI@home project\n"),
	_("page, can be found in the config panel under the User Info tab.\n\n"),
	_("<b>Theme:\n\n"),
	_("You can theme you SETI@home progress meter by simply creating\n"),
	_("a "),
	_("<i>seti"),
	_(" directory under your favorite theme directory.\n\n"),
	_("There, you have to put some files:\n  -"),
	_("<i>bg_meter.ext:"),
	_(" the background of the meter,\n  -"),
	_("<i>krell.ext:"),
	_(" used to replace the flying saucer in this meter,\n  -"),
	_("<i>bg_meter_pcpu.ext:"),
	_(" the background of the CPU meter when displayed, and\n  -"),
	_("<i>krell_pcpu.ext:"),
	_(" used to replace the CPU krell when displayed.\n\n"),
	_("Where "),
	_("<i>.ext"),
	_(" represents a GKrellM understandable file format like png or xpm.\n\n"),
	_("If you use new krells, you'll have to put information in your\n"),
	_("gkrellmrc file. The main panel is configured via usual gkrellm commands.\n"),
	_("The CPU usage krell can be configured with the following commands.\n"),
	_("<i>The krell:\n"),
	_("  set_integer seti_pcpu_x_hot = xval\n"),
	_("  set_integer seti_pcpu_yoff = yval\n"),
	_("  set_integer seti_pcpu_depth = dval\n\n"),
	_("<i>The background:\n"),
	_("  set_image_border seti_pcpu_bg_meter = l,r,t,b\n"),
	_("  set_integer seti_pcpu_top_margin = tm\n"),
	_("  set_integer seti_pcpu_bottom_margin = bm\n\n"),
	_("You can also change the properties of the label by adding lines to\n"),
	_("your gkrellmrc file. The keyword is \"seti\" (take a look at the theme doc).\n")
};

static gchar seti_about_text[] =
"SETI@home GKrellM plugin version " MAJOR_VERSION "." MINOR_VERSION "." RELEASE_NUMBER"\n\n" \
"Copyright (C) 2000 Xavier Serpaggi and Tilman Sauerbeck\n" \
"xavier.serpaggi@free.fr   tilman.sauerbeck@web.de\n" \
"http://xavier.serpaggi.free.fr/seti/\n\n" \
"SETI@home GKrellM plugin comes with ABSOLUTELY NO WARRANTY.\n" \
"This is free software, and you are welcome to redistribute\n" \
"it under certain conditions.\n" \
"See the GNU Public License for details" ;


/** cycle through the modes used to display CPU usage
 */
static void set_pcpu_mode()
{
	switch ( pcpu_display ) {
		case 0:
			if (gkrellm_is_panel_visible(panel_pcpu) == TRUE)
				gkrellm_panel_hide(panel_pcpu);
			if (krell_inserted == FALSE)
				gkrellm_insert_krell(panel, k_pcpu2, 0);
			krell_inserted = TRUE;
			break;
		case 1:
			if (gkrellm_is_panel_visible(panel_pcpu) == FALSE)
				gkrellm_panel_show(panel_pcpu);
			gkrellm_remove_krell(panel, k_pcpu2);
			krell_inserted = FALSE;
			break;
		case 2:
			if (gkrellm_is_panel_visible(panel_pcpu) == TRUE)
				gkrellm_panel_hide(panel_pcpu);
			gkrellm_remove_krell(panel, k_pcpu2);
			krell_inserted = FALSE;
			break;
	}
}


/** Update the tooltip.
 * 
 * The tooltip won't disapear anymore when a minute is over.
 * client_info.state is used to notify wether the client
 * is running or not.
 *
 * \warning This one does not work. Please, help!
 */
static void update_tooltip(void)
{
	gint	 nresults = 1 ;
	
	if ( ! seti_paths_status ) {
		g_snprintf(tooltip_text, 26, _("It seems there is a problem with the specified path!")) ;
	} else {
		nresults += atoi(user_info.nresults) ;

		if ( client_info.state )
			g_snprintf(tooltip_text, 52, _("Workunit %d - %.2f%% complete (%s)"), nresults, work_unit_info.lprogress, seti_status_report[work_unit_info.status]) ;
		else
			g_snprintf(tooltip_text, 52, _("Workunit %d - %.2f%% complete (%s-idle)"), nresults, work_unit_info.lprogress, seti_status_report[work_unit_info.status]) ;
	}
	gtk_tooltips_set_tip( tooltip, GTK_WIDGET(panel->drawing_area), tooltip_text , _("SETI@home gkrellm plugin: client status")) ;
	gtk_tooltips_enable (tooltip);
}


/** The main function, called by GKrellM to update the * krell.
 *
 * Big checks are done only every minute.
 */
static void update_seti()
{
	static gint	 tic_toc = 0 ;
	static gint	 previous_state = -1 ;
	gint		 work_unit_previous_status = -1 ;
	gint		 pid ;


	if ( ! seti_paths_status ) return ;
	
	/* Things to do when the client is running. */
	if ( seti_is_running(&pid) ) {

		tic_toc = (0 == tic_toc) ;

		/* If the client has just been restarted (state == running and state != previous_state) (force check),
		 * or it is a minute tick.
		 */
		if ( previous_state != client_info.state || GK.minute_tick ) {
			if ( previous_state != client_info.state ) {
				previous_state = client_info.state ;
				clean_user_info() ;
				get_user_info() ;
#ifdef DEBUG
				if ( GK.debug ) printf(_("SETI@home plugin: client restarted\n")) ;
#endif
			}

			get_work_unit_status() ;
			if ( work_unit_previous_status > work_unit_info.status ) work_unit_info.lprogress = 0.0 ;
			work_unit_previous_status = work_unit_info.status ;
			
			/* GKrellM updates routines are designed to record differences between
			 |  the current data value and the data value from the previous call.
			 |  This is for handling continuously increasing data values.  However,
			 |  if you want to measure values between a fixed range, you can do so
			 |  by setting previous to zero before each update call.
			 */
			k->previous = 0;
			k->full_scale = 100;
			gkrellm_update_krell(panel, k, work_unit_info.lprogress );

			/* Updating a krell (or drawing a decal) draws only into local
			 |  pixmaps.  After all krells and decals are "drawn", they must be
			 |  really drawn to the screen with gkrellm_draw_panel_layers().
			 */
			gkrellm_draw_panel_layers(panel);
			
			update_tooltip() ;
		}
	
		if (GK.second_tick) {
			work_unit_info.pcpu = get_pcpu_info(pid);
			if (work_unit_info.pcpu < 0)
				work_unit_info.pcpu = 0;
				
			k_pcpu->previous = 0;
			k_pcpu->full_scale = 100;
			gkrellm_update_krell(panel_pcpu, k_pcpu, work_unit_info.pcpu);
			
			k_pcpu2->previous = 0;
			k_pcpu2->full_scale = 100;
			gkrellm_update_krell(panel, k_pcpu2, work_unit_info.pcpu);
			
			gkrellm_draw_panel_layers(panel);
		
			g_snprintf(pcpu_text, 10, "%2.1f%% CPU", work_unit_info.pcpu);
			gkrellm_draw_decal_text(panel_pcpu, decal_text_pcpu, pcpu_text, -1);
			gkrellm_draw_panel_layers(panel_pcpu);
		}
		
	} else {
		if ( GK.second_tick ) {
			if ( work_unit_info.lprogress > 99.0 && work_unit_info.lprogress < 100.0 ) {
				get_work_unit_status() ;
				if ( work_unit_previous_status > work_unit_info.status ) work_unit_info.lprogress = 0.0 ;
				work_unit_previous_status = work_unit_info.status ;
			}

			/* I guess this is not a really good solution to make the
			 * flying saucer blink. Give me a better one please.
			 */
			tic_toc = (0 == tic_toc) ;      

			k->previous = 0;
			k->full_scale = 100;

			gkrellm_update_krell(panel, k, (gdouble)tic_toc * work_unit_info.lprogress) ;
			
			/* The client doesn't run, so no pcpu. */
			k_pcpu->previous = 0;
			k_pcpu->full_scale = 100;
			gkrellm_update_krell(panel_pcpu, k_pcpu, 0);
			
			k_pcpu2->previous = 0;
			k_pcpu2->full_scale = 100;
			gkrellm_update_krell(panel, k_pcpu2, 0);
			
			gkrellm_draw_panel_layers(panel);
			
			strncpy(pcpu_text, "0.0% CPU  ", 10);
			gkrellm_draw_decal_text(panel_pcpu, decal_text_pcpu, pcpu_text, -1);
			gkrellm_draw_panel_layers(panel_pcpu);
		}

		/* The client has just been stopped. */
		if ( previous_state != client_info.state ) {
			previous_state = client_info.state ;
			clean_user_info() ;
			get_user_info() ;
#ifdef DEBUG
			if ( GK.debug ) printf(_("SETI@home plugin: client stopped\n")) ;
#endif
			update_tooltip() ;
		}
	}
}


static gint panel_expose_event(GtkWidget *widget, GdkEventExpose *ev)
{
	gdk_draw_pixmap(widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			panel->pixmap, ev->area.x, ev->area.y, ev->area.x, ev->area.y,
			ev->area.width, ev->area.height);
	return FALSE;
}

static gint panel_pcpu_expose_event(GtkWidget *widget, GdkEventExpose *ev)
{
	gdk_draw_pixmap(widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			panel_pcpu->pixmap, ev->area.x, ev->area.y, ev->area.x, ev->area.y,
			ev->area.width, ev->area.height);
	return FALSE;
}



/* User has clicked panel. We should try to start/stop seti@home software
 * client_stop() and client_start() inspired by e-setimon from Ben Wilson.
 */
static void client_stop(gint pid)
{
	int	res ;
	char	runthis[256];


	if ( -1 == pid ) {
		system ("killall setiathome");
	} else if ( strcmp(client_info.seti_client_stop, "internal") == 0 ) {
		/* This solution avoid to do a `system' call, and
		 | thus is less exposed to security issues if run
		 | by root. But who wants to run GKrellM as root!
		 */
		res = kill(pid, SIGTERM) ;
#ifdef DEBUG
		if ( (res == -1) && (GK.debug) ) printf(_("SETI@home plugin: unable to kill setiathome client pid %d\n"), pid) ;
#endif
	} else {
		sprintf (runthis, "%s\n", client_info.seti_client_stop); 
		system (runthis);
#ifdef DEBUG
		if ( GK.debug ) printf(_("SETI@home plugin: killing the client with provided script : %s\n"), client_info.seti_client_stop) ;
#endif
	}
}


static void client_start(void)
{
	char runthis[256];


	if ( strcmp(client_info.seti_client_start, "internal") == 0 ) {
		sprintf (runthis, "cd %s ; %s/setiathome %s & \n", client_info.seti_file_path, client_info.seti_client_path, client_info.seti_args); 

#ifdef DEBUG
		if ( GK.debug ) printf(_("SETI@home plugin: command passed to client = \"%s\"\n"), runthis) ;
#endif

	} else {
		sprintf (runthis, "%s\n", client_info.seti_client_start); 

#ifdef DEBUG
		if ( GK.debug ) printf(_("SETI@home plugin: launching the client with provided script : %s\n"), client_info.seti_client_start) ;
#endif
	}

	system (runthis);
}


static gint panel_click_event(GtkWidget *widget, GdkEventButton *ev)
{
	gint pid ;


	if ( 1 == ev->button ) {
#ifdef DEBUG
		if ( GK.debug ) printf(_("Left click on SETI@home plugin: will try to start/stop client\n"));
#endif
		if ( seti_is_running(&pid) ) {
			client_stop(pid) ;
			/* clean_user_info() ; */
		} else if ( -1 == pid ) {
			/* get_user_info() ; */
			client_start() ;
		}
	} else if ( 2 == ev->button ) {
#ifdef DEBUG
		if ( GK.debug ) printf(_("Middle click on SETI@home plugin\n"));
#endif
		pcpu_display = (pcpu_display + 1) % 3 ;
		set_pcpu_mode();
		gkrellm_config_modified();
	} else if ( 3 == ev->button ) {
		gkrellm_open_config_window(monitor);
#ifdef DEBUG
		if ( GK.debug ) printf(_("Right click on SETI@home plugin\n"));
#endif
	}

	return 1 ;
}


static gint panel_pcpu_click_event(GtkWidget *widget, GdkEventButton *ev)
{
	if ( 1 == ev->button ) {
#ifdef DEBUG
		if ( GK.debug ) printf(_("Left click on SETI@home plugin (CPU usage)\n"));
#endif
	} else if ( 2 == ev->button ) {
#ifdef DEBUG
		if ( GK.debug ) printf(_("Middle click on SETI@home plugin (CPU usage)\n"));
#endif
		pcpu_display = (pcpu_display + 1) % 3 ;
		set_pcpu_mode();
		gkrellm_config_modified();
	} else if ( 3 == ev->button ) {
		gkrellm_open_config_window(monitor);
#ifdef DEBUG
		if ( GK.debug ) printf(_("Right click on SETI@home plugin (CPU usage)\n"));
#endif
	}

	return 1 ;
}


static void create_seti(GtkWidget *vbox, gint first_create)
{
	GkrellmStyle		*style_seti = NULL ;
	GkrellmStyle		*style_pcpu = NULL ;
	GkrellmTextstyle	*ts = NULL ;

	
	if (first_create) {
		panel = gkrellm_panel_new0();
		panel_pcpu = gkrellm_panel_new0();
	}

	style_seti = gkrellm_meter_style(style_id) ;
	style_seti->krell_yoff = 3 ;
	style_seti->krell_depth = 22 ;
	style_seti->krell_x_hot = 24 ;
	gkrellm_load_piximage("krell_seti", krell_seti_xpm, &krell_image_seti, STYLE_NAME) ;
	gkrellm_load_piximage("bg_meter_seti", bg_meter_seti_xpm, &bg_meter_image_seti, STYLE_NAME) ;
	gkrellm_get_gkrellmrc_piximage_border("seti_bg_meter", bg_meter_image_seti, &style_seti->border) ;
	/*
	gkrellm_get_gkrellmrc_integer("seti_top_margin", &style_seti->top_margin) ;
	gkrellm_get_gkrellmrc_integer("seti_bottom_margin", &style_seti->bottom_margin) ;
	*/
	gkrellm_panel_bg_piximage_override(panel, bg_meter_image_seti) ; 

	k = gkrellm_create_krell(panel, krell_image_seti, style_seti);
	gkrellm_monotonic_krell_values(k, FALSE);
	gkrellm_set_krell_full_scale(k, 30, 1);

	/* No need to do that as it is about the main panel.
	 * Things are themed using the classical GKrellM way in gkrellmrc.
	gkrellm_get_gkrellmrc_integer("seti_x_hot", &style_seti->krell_x_hot);
	gkrellm_get_gkrellmrc_integer("seti_yoff", &style_seti->krell_yoff);
	gkrellm_get_gkrellmrc_integer("seti_depth", &style_seti->krell_depth);
	*/	
	
	
	/* The second panel style is a copy of the main panel unless it is themed. */
	if (style_pcpu) g_free(style_pcpu) ;
	style_pcpu = gkrellm_copy_style(style_seti) ;

	if ( bg_meter_image_pcpu ) {
		gkrellm_destroy_piximage(bg_meter_image_pcpu);
		bg_meter_image_pcpu = NULL;
	}
	gkrellm_load_piximage("bg_meter_pcpu", NULL, &bg_meter_image_pcpu, STYLE_NAME);
			
	if ( bg_meter_image_pcpu ) {
		gkrellm_get_gkrellmrc_piximage_border("seti_pcpu_bg_meter", bg_meter_image_pcpu, &style_pcpu->border);
		gkrellm_panel_bg_piximage_override(panel_pcpu, bg_meter_image_pcpu);
	}
	/*
	gkrellm_get_gkrellmrc_integer("seti_pcpu_top_margin", &style_pcpu->top_margin);
	gkrellm_get_gkrellmrc_integer("seti_pcpu_bottom_margin", &style_pcpu->bottom_margin);
	*/

	style_pcpu->krell_yoff = 4 ;
	style_pcpu->krell_depth = 1 ;
	style_pcpu->krell_x_hot = -1 ;
	gkrellm_get_gkrellmrc_integer("seti_pcpu_yoff", &style_pcpu->krell_yoff);
	gkrellm_get_gkrellmrc_integer("seti_pcpu_depth", &style_pcpu->krell_depth);
	gkrellm_get_gkrellmrc_integer("seti_pcpu_x_hot", &style_pcpu->krell_x_hot);
	gkrellm_load_piximage("krell_pcpu", krell_pcpu_xpm, &krell_image_pcpu, STYLE_NAME) ;

	/* krell in the second panel: */
	k_pcpu = gkrellm_create_krell(panel_pcpu, krell_image_pcpu, style_pcpu);
	gkrellm_monotonic_krell_values(k_pcpu, FALSE);
	gkrellm_set_krell_full_scale(k_pcpu, 30, 1);
	
	/* krell in seti panel: */
	k_pcpu2 = gkrellm_create_krell(panel, krell_image_pcpu, style_pcpu);
	gkrellm_monotonic_krell_values(k_pcpu2, FALSE);
	gkrellm_set_krell_full_scale(k_pcpu2, 30, 1);
	
	ts = gkrellm_meter_textstyle(style_id);
	decal_text_pcpu = gkrellm_create_decal_text(panel_pcpu, "CPU4SETI", ts, style_pcpu, -1, -1, -1);
	
	/* Configure panel calculates the panel height needed for the "Plugin" label.
	 |  and the krell.
	 */
	panel->textstyle = gkrellm_meter_textstyle(style_id);
	gkrellm_panel_configure(panel, "SETI", style_seti);
	
	panel_pcpu->textstyle = gkrellm_meter_textstyle(style_id);
	gkrellm_panel_configure(panel_pcpu, NULL, style_pcpu);

	/* Build the configured panel with a background image and pack it into
	 |  the vbox assigned to this monitor.  The background image will be
	 |  the theme default or plugin specific if created by the theme author.
	 */
	gkrellm_panel_create(vbox, monitor, panel);
	gkrellm_panel_create(vbox, monitor, panel_pcpu);
	

	if (first_create) {
		gtk_signal_connect(GTK_OBJECT (panel->drawing_area), "expose_event",
				(GtkSignalFunc) panel_expose_event, NULL);
		gtk_signal_connect(GTK_OBJECT (panel->drawing_area), "button_release_event",
				(GtkSignalFunc) panel_click_event, NULL);
		
		gtk_signal_connect(GTK_OBJECT (panel_pcpu->drawing_area), "expose_event",
				(GtkSignalFunc) panel_pcpu_expose_event, NULL);
		gtk_signal_connect(GTK_OBJECT (panel_pcpu->drawing_area), "button_release_event",
				(GtkSignalFunc) panel_pcpu_click_event, NULL);

		get_user_info() ;
		get_work_unit_status() ;
	}
	
	if ( tooltip == NULL ) {
		tooltip = gtk_tooltips_new() ;
		gtk_tooltips_set_delay(tooltip, 500);
		update_tooltip() ;
	}
	
	krell_inserted = TRUE;
	set_pcpu_mode();
}


static void create_seti_config(GtkWidget *tab_vbox)
{
	GtkWidget       *frame;
	GtkWidget       *vbox;
	GtkWidget       *hbox;
	GtkWidget       *hbox1;
	GtkWidget       *tabs;
	GtkWidget       *text;
	GtkWidget       *label;
	GtkWidget       *info_label;
	GtkWidget       *scrolled;
	GSList		*group;
	GtkWidget	*button;


	tabs = gtk_notebook_new() ;
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP) ;
	gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0) ;

	/* Config tab */
	vbox = gkrellm_gtk_framed_notebook_page(tabs, _("Config")) ;
	
	label = gtk_label_new(_("Path to state.sah file:")) ;
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;

	seti_path_entry = gtk_entry_new_with_max_length(255) ;
	gtk_entry_set_text(GTK_ENTRY(seti_path_entry), client_info.seti_file_path) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_path_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_path_entry, FALSE, FALSE, 0) ;

	label = gtk_label_new(_("Path to seti client:"));
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;

	seti_client_entry = gtk_entry_new_with_max_length(255) ;
	gtk_entry_set_text(GTK_ENTRY(seti_client_entry), client_info.seti_client_path) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_client_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_client_entry, FALSE, FALSE, 0) ;

	label = gtk_label_new(_("Arguments to the client:")) ;
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;

	seti_args_entry = gtk_entry_new_with_max_length(255);
	gtk_entry_set_text(GTK_ENTRY(seti_args_entry), client_info.seti_args) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_args_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_args_entry, FALSE, FALSE, 0) ;
	

	/* Client control tab */
	vbox = gkrellm_gtk_framed_notebook_page(tabs, _("Client Control")) ;

	label = gtk_label_new(_("Command to start the client:")) ;
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;

	seti_client_start_entry = gtk_entry_new_with_max_length(255);
	gtk_entry_set_text(GTK_ENTRY(seti_client_start_entry), client_info.seti_client_start) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_client_start_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_client_start_entry, FALSE, FALSE, 0) ;

	label = gtk_label_new(_("Command to stop the client:")) ;
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;

	seti_client_stop_entry = gtk_entry_new_with_max_length(255);
	gtk_entry_set_text(GTK_ENTRY(seti_client_stop_entry), client_info.seti_client_stop) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_client_stop_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_client_stop_entry, FALSE, FALSE, 0) ;
	
	
	/* CPU Usage tab */
	vbox = gkrellm_gtk_framed_notebook_page(tabs, _("CPU Usage"));

	label = gtk_label_new(_("How do you want CPU usage info to be displayed?"));
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;

	button = gtk_radio_button_new_with_label(NULL, _(_("Show CPU usage with a second krell in the SETI panel")));
	gtk_box_pack_start(GTK_BOX (vbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	pcpu_select_button[0] = button;
	button = gtk_radio_button_new_with_label(group, _(_("Show CPU usage in a second panel")));
	gtk_box_pack_start(GTK_BOX(vbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	pcpu_select_button[1] = button;
	button = gtk_radio_button_new_with_label(group, _(_("Don't show CPU usage info")));
	gtk_box_pack_start(GTK_BOX(vbox), button, TRUE, TRUE, 0);
	group = gtk_radio_button_group(GTK_RADIO_BUTTON (button));
	pcpu_select_button[2] = button;
	
	button = pcpu_select_button[pcpu_display];
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);


	/* User info tab */
	vbox = gkrellm_gtk_framed_notebook_page(tabs, _("User Info")) ;

	hbox = gtk_hbox_new(TRUE, 0) ;
	hbox1 = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Name: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.name) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, TRUE, TRUE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(hbox), hbox1, TRUE, TRUE, 0 ) ;
	
	hbox1 = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Work Unit Sent: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.nresults) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, TRUE, TRUE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(hbox), hbox1, TRUE, TRUE, 0 ) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0 ) ;

	hbox = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Email address: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.email_addr) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0 ) ;

	hbox = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("URL: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.url) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0 ) ;

	hbox = gtk_hbox_new(TRUE, 0) ;
	hbox1 = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Country: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.country) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, TRUE, TRUE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(hbox), hbox1, TRUE, TRUE, 0 ) ;

	hbox1 = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Postal Code: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.postal_code) ;
	gtk_box_pack_start(GTK_BOX(hbox1), label, TRUE, TRUE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(hbox), hbox1, TRUE, TRUE, 0 ) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0 ) ;

	hbox = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Register Time: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.register_time) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0 ) ;

	/*
	hbox = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Last WU Time: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.last_wu_time) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0 ) ;
	*/
	
	hbox = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Last Result Time: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.last_result_time) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0 ) ;
	
	hbox = gtk_hbox_new(FALSE, 0) ;
	label = gtk_label_new(_("Total CPU Time: ")) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	label = gtk_label_new(user_info.total_cpu) ;
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0 ) ;

	/* Info tab */
	frame = gtk_frame_new(NULL) ;
	gtk_container_border_width(GTK_CONTAINER(frame), 3) ;

	label = gtk_label_new(_("Info"));
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs), frame, label);

	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(frame), scrolled) ;

	text = gtk_text_view_new();

	gkrellm_gtk_text_view_append_strings(text, seti_info_text, sizeof(seti_info_text) / sizeof(gchar *));
	gtk_text_view_set_editable(GTK_TEXT_VIEW(text), FALSE);
	gtk_container_add(GTK_CONTAINER(scrolled), text);


	/* About tab */
	info_label = gtk_label_new(seti_about_text);
	label = gtk_label_new(_("About"));
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs), info_label, label);
}


/*
 *      Saving config
 */
static void save_seti_config(FILE *f)
{
	fprintf(f, _("seti filepath %s\n"), client_info.seti_file_path) ;
	fprintf(f, "seti args %s\n", client_info.seti_args) ;
	fprintf(f, "seti clientpath %s\n", client_info.seti_client_path) ;
	fprintf(f, "seti client_start %s\n", client_info.seti_client_start) ;
	fprintf(f, "seti client_stop %s\n", client_info.seti_client_stop) ;
	fprintf(f, "seti pcpu %i\n", pcpu_display) ;
#ifdef DEBUG
	if ( GK.debug ) printf (_("SETI@home plugin save:\n\t%s\n\t%s\n"), client_info.seti_file_path, client_info.seti_args) ;
#endif

	(void) seti_paths_exist() ;
}


/*
 *  Loading config
 */
static void load_seti_config(gchar *arg)
{
	gchar   seti_config[64], item[256];
	gint    n;

	n = sscanf(arg, "%s %[^\n]", seti_config, item);


	if (n == 2) {
#ifdef DEBUG
		if ( GK.debug ) printf(_("SETI@home plugin config: %s item=\"%s\"\n"), seti_config, item) ;
#endif
		if (strcmp(seti_config, "filepath") == 0) {

			g_free(client_info.seti_file_path) ;
			client_info.seti_file_path = g_strdup(item) ;

			g_free(client_info.state_file) ;
			client_info.state_file = g_strdup_printf("%s/state.sah", client_info.seti_file_path) ;

			g_free(client_info.user_file) ;
			client_info.user_file = g_strdup_printf("%s/user_info.sah", client_info.seti_file_path) ;
#ifdef DEBUG
			if ( GK.debug ) printf (_("SETI@home plugin load:\n\t%s\n\t%s\n\t%s\n"), client_info.seti_file_path, client_info.state_file, client_info.user_file) ;
#endif
		} else if (strcmp(seti_config, "args") == 0) {

			g_free(client_info.seti_args) ;
			client_info.seti_args = g_strdup(item) ;
#ifdef DEBUG
			if ( GK.debug ) printf (_("SETI@home plugin load:\n\t%s\n"), client_info.seti_args) ;
#endif
		} else if (strcmp(seti_config, "clientpath") == 0) {

			g_free(client_info.seti_client_path) ;
			client_info.seti_client_path = g_strdup(item) ; 
#ifdef DEBUG
			if ( GK.debug ) printf (_("SETI@home plugin load:\n\t%s\n"), client_info.seti_client_path) ;
#endif
		} else if (strcmp(seti_config, "client_start") == 0) {
			g_free(client_info.seti_client_start) ;
			client_info.seti_client_start = g_strdup(item) ;
#ifdef DEBUG
			if ( GK.debug ) printf (_("SETI@home plugin load:\n\t%s\n"), client_info.seti_client_start) ;
#endif
		} else if (strcmp(seti_config, "client_stop") == 0) {
			g_free(client_info.seti_client_stop) ;
			client_info.seti_client_stop = g_strdup(item) ;
#ifdef DEBUG
			if ( GK.debug ) printf (_("SETI@home plugin load:\n\t%s\n"), client_info.seti_client_stop) ;
#endif
		} else if (strcmp(seti_config, "pcpu") == 0) {
			pcpu_display = 0;
			pcpu_display = atoi(item) ;
#ifdef DEBUG
			if ( GK.debug ) printf (_("SETI@home plugin load:\n\t%i\n"), pcpu_display) ;
#endif
		}
	}

	(void) seti_paths_exist() ;
	clean_user_info() ;
	get_user_info() ;
}


/*
 *      Applying config.
 */
static void apply_seti_config()
{
	static gint	i;
	
	g_free(client_info.seti_file_path) ;
	client_info.seti_file_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_path_entry))) ;

	g_free(client_info.seti_client_path) ;
	client_info.seti_client_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_client_entry))) ;

	g_free(client_info.state_file) ;
	client_info.state_file = g_strdup_printf("%s/state.sah", client_info.seti_file_path) ;

	g_free(client_info.user_file) ;
	client_info.user_file = g_strdup_printf("%s/user_info.sah", client_info.seti_file_path) ;

	g_free(client_info.seti_args) ;
	client_info.seti_args = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_args_entry))) ;

	g_free(client_info.seti_client_start) ;
	client_info.seti_client_start = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_client_start_entry))) ;

	g_free(client_info.seti_client_stop) ;
	client_info.seti_client_stop = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_client_stop_entry))) ;
	
	for (i=0; i<3; i++)
		if (GTK_TOGGLE_BUTTON(pcpu_select_button[i])->active)
			pcpu_display = i;
	set_pcpu_mode();
		
	(void) seti_paths_exist() ;
	clean_user_info() ;
	get_user_info() ;

#ifdef DEBUG
	if ( GK.debug ) printf(_("SETI@home plugin apply:\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n\t%s\n"), client_info.seti_file_path, client_info.seti_client_path, client_info.state_file, client_info.user_file, client_info.seti_args, client_info.seti_client_start, client_info.seti_client_stop) ;
#endif
}


/* The monitor structure tells GKrellM how to call the plugin routines.  */
static GkrellmMonitor  plugin_mon      =
{
	CONFIG_NAME,	    /* Title for config clist.  */
	0,		    /* Id, 0 if a plugin        */
	create_seti,	    /* The create function      */
	update_seti,	    /* The update function      */


	create_seti_config, /* The config tab create function   */
	apply_seti_config,  /* Apply the config function	*/


	save_seti_config,   /* Save user config		     */
	load_seti_config,   /* Load user config		     */
	"seti",		    /* config keyword		     */


	NULL,		    /* Undefined 2  */
	NULL,		    /* Undefined 1  */
	NULL,		    /* private      */


	SETI_PLACE,         /* Insert plugin before this monitor */


	NULL,		    /* Handle if a plugin, filled in by GKrellM     */
	NULL		    /* path if a plugin, filled in by GKrellM       */
};


/* All GKrellM plugins must have one global routine named gkrellm_init_plugin()
 * which returns a pointer to a filled in monitor structure.
 */
GkrellmMonitor *gkrellm_init_plugin()
{
	/* If this next call is made, the background and krell images for this
	 |  plugin can be custom themed by putting bg_meter.png or krell.png in the
	 |  subdirectory STYLE_NAME of the theme directory.  Text colors (and
	 |  other things) can also be specified for the plugin with gkrellmrc
	 |  lines like:  StyleMeter  STYLE_NAME.textcolor orange black shadow
	 |  If no custom theming has been done, then all above calls using
	 |  style_id will be equivalent to style_id = DEFAULT_STYLE_ID.
	 */
	style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);

	g_free(client_info.seti_file_path) ;
	client_info.seti_file_path = g_strdup("/opt/setiathome/") ;

	g_free(client_info.seti_client_path) ;
	client_info.seti_client_path = g_strdup("/opt/setiathome/") ;

	g_free(client_info.state_file) ;
	client_info.state_file = g_strdup_printf("%s/state.sah", client_info.seti_file_path) ;

	g_free(client_info.user_file) ;
	client_info.user_file = g_strdup_printf("%s/user_info.sah", client_info.seti_file_path) ;

	g_free(client_info.seti_args) ;
	client_info.seti_args = g_strdup("-nice 19 -email > /dev/null 2> /dev/null") ;

	g_free(client_info.seti_client_start) ;
	client_info.seti_client_start = g_strdup("internal") ;

	g_free(client_info.seti_client_stop) ;
	client_info.seti_client_stop = g_strdup("internal") ;

	tooltip = NULL ;

	seti_paths_status = 0 ;
	
	/* Get the current locale, as it seems sscanf wants it */
	locale_conv = localeconv() ;
	
	pcpu_display = 0;
	my_switch = 0;
	strncpy(pcpu_text, "0.0%% CPU  ", 10);
	krell_inserted = FALSE;

#ifdef DEBUG
	if ( GK.debug && locale_conv ) printf(_("On this system the decimal point is [%s]\n"), locale_conv->decimal_point) ;
#endif

	monitor = &plugin_mon;
	return &plugin_mon;
}
