/* SETI@home plugin for GKrellM
 *
 *  Copyright (c) 2000 Xavier Serpaggi
 *
 *  Author: Xavier Serpaggi <xavier dot serpaggi at free dot fr>
 *
 *  Version: 0.7.0
 *
 *  This program is free software which I release under the GNU General Public
 *  License. You may redistribute and/or modify this program under the terms
 *  of that license as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
/* seti.h - see seti.c for details on how to compile this program.
 */
 

#include <gkrellm2/gkrellm.h>
#if (GKRELLM_VERSION_MAJOR<1) || (GKRELLM_VERSION_MAJOR==1 \
  && GKRELLM_VERSION_MINOR==2 && GKRELLM_VERSION_REV<2)
#error This plugin requires GKrellM version >= 1.2.2
#endif

#include "pixmaps/bg_meter_seti.xpm"
#include "pixmaps/krell_seti.xpm"
#include "pixmaps/krell_pcpu.xpm"


#define CONFIG_NAME     "SETI@home"     /* Name in the configuration window. */
#define STYLE_NAME      "seti"	  	/* Theme subdirectory name and gkrellmrc style name. */
#define MAJOR_VERSION	"0"
#define MINOR_VERSION	"7"
#define	RELEASE_NUMBER	"0b"

/* Added by Tilman Sauerbeck, Aug 30. 100 seems to be correct for i386 systems.
 * You have to set the right value for your platform here, or the pcpu value won't be correct
 */
#define HERTZ 100


/**
 *  The location of the plugin. You can choose from:
 *   MON_CLOCK, MON_CPU, MON_PROC, MON_DISK,
 *   MON_INET, MON_NET, MON_FS, MON_MAIL,
 *   MON_APM, or MON_UPTIME
 *
 *  The plugin will place itself above SETI_PLACE
 *
 *  The gravity defines where the plugin should be placed in comparison
 *  with other plugins with the same placement position.
 *  If you want the plugin higher (but with the same placement), you
 *  set the gravity a bit lower, and if you want the plugin lower (with
 *  the same placement), you should set the gravity a bit higher.
 *  Min is 0, max is 15.
 *
 */
#ifdef GRAVITY
#	define SETI_PLACE (MON_APM | GRAVITY(10))
#else
#	define SETI_PLACE (MON_APM)
#endif
/** Information about the user gathered from the SETI@home project site.
 */
typedef struct {
	gchar	*email_addr ;
	gchar	*name ;
	gchar	*url ;
	gchar	*country ;
	gchar	*postal_code ;
	gchar	*register_time ;
	gchar	*last_wu_time ;
	gchar	*last_result_time ;
	gchar	*total_cpu ;
	gchar	*nresults ;
} T_user_info, *P_user_info ;

/** Information about the client been monitored.
 */
typedef struct {
	gchar	*seti_file_path ;
	gchar	*state_file ;
	gchar	*user_file ;
	gchar	*seti_args ;
	gchar	*seti_client_path ;
	gchar	*seti_client_start ;
	gchar	*seti_client_stop ;
	gint	 state ;
} T_client_info, *P_client_info ;

/** Information about the workunit itself.
 */
typedef struct {
	gdouble	 lprogress ;
	gint	 status ;
	gdouble	 pcpu ;
} T_work_unit_info, *P_work_unit_info ;
