#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <locale.h>

#include "seti.h"
#include "utils.h"


extern gint seti_paths_status ;
extern struct stat seti_stat_buf ;
extern struct lconv *locale_conv ;
extern T_user_info user_info ;
extern T_client_info client_info ;
extern T_work_unit_info work_unit_info ;
extern gchar buf[256] ;
extern gdouble total_time[2] ;
extern gboolean my_switch ;
extern gdouble seconds[2] ;


/** Check if the paths specified (or not yet!) in the config tab actually exist.
 * 
 * Return 0 if they do not exist.
 */
gint seti_paths_exist(void)
{
	seti_paths_status = 0 ;

	/* First, check if the path to the client is correct. */
	if ( stat(client_info.seti_client_path, &seti_stat_buf) != 0 ) {
#ifdef DEBUG
		if ( GK.debug ) printf("SETI@home plugin: path to client does not exist.\n") ;
#endif
		return 0 ;
	}

	if ( ! S_ISDIR(seti_stat_buf.st_mode) ) {
#ifdef DEBUG
		if ( GK.debug ) printf("SETI@home plugin: path to client does not exist.\n") ;
#endif
		return 0 ;
	}

#ifdef DEBUG
		if ( GK.debug ) printf("SETI@home plugin: path to client found!\n") ;
#endif


	/* Then, check if the path to the workunit is correct. */
	if ( stat(client_info.seti_file_path, &seti_stat_buf) != 0 ) {
#ifdef DEBUG
		if ( GK.debug ) printf("SETI@home plugin: path to workunit does not exist.\n") ;
#endif
		return 0 ;
	}

	if ( ! S_ISDIR(seti_stat_buf.st_mode) ) {
#ifdef DEBUG
		if ( GK.debug ) printf("SETI@home plugin: path to workunit does not exist.\n") ;
#endif
		return 0 ;
	}

#ifdef DEBUG
		if ( GK.debug ) printf("SETI@home plugin: path to workunit found!\n") ;
#endif

	
	return seti_paths_status = 1 ;
}

/** Check is there's a process with pid number equal to the one found int pid.sah file of seti client.
 *
 * returns 0 if not running.
 */
gint seti_is_running(gint *pid)
{
	FILE	*f ;
	gchar 	*pid_file,
		 pid_string[10],
		 status='?',
		 long_status[255] = "unknown" ;


	if ( ! seti_paths_status ) return 0 ;
	
	pid_file = g_strdup_printf("%s/pid.sah", client_info.seti_file_path) ;
	if ( !(f = fopen(pid_file, "r")) ) {
		g_free(pid_file) ;
		*pid = -1 ;
		client_info.state = 0 ;
		return 0 ;
	}
	fgets(pid_string, 10, f) ;
	fclose (f) ;
	g_free(pid_file) ;

	if ( sscanf(pid_string, "%d", pid) == 0 ) {
		*pid = -1 ;
		client_info.state = 0 ;
		return 0 ;
	}

	pid_file = g_strdup_printf("/proc/%d/status", *pid) ;
	if ( !(f = fopen(pid_file, "r")) ) {
		g_free(pid_file) ;
		*pid = -1 ;
		client_info.state = 0 ;
		return 0 ;
	} else {
		while ( (fgets(buf, sizeof(buf), f)) != NULL ) {
			if ( strncmp(buf, "State:", 6) == 0 ) {
				sscanf(buf, "State:  %c (%[^()])%*s", &status,  long_status) ;
			}
		}
		fclose(f) ;
		g_free(pid_file) ;
		
		if ( status != 'R' ) {
			*pid = -1 ;
			client_info.state = 0 ;
			return 0 ;
		}
	}

	client_info.state = 1 ;
	return 1 ;
}

/** translate the seconds given by SETI@home server to a human redeable form
 *  (looks like on the seti home page).
 */
gchar *strtotime(char *Seconds)
{
	gdouble	 temp ;
	gint	 years, days, hours, minutes, seconds ;
	
	temp	= strtod(Seconds, NULL) ;

	years	= (gint) (temp / 3600. / 24. / 365.25) ;
	days	= (gint) (temp / 3600. / 24. - years * 365.25) ;
	hours	= (gint) (temp / 3600. - years * 365.25 * 24. - days * 24.) ;
	minutes	= (gint) (temp / 60 - years * 365.25 * 24. * 60. - days * 24. * 60. - hours * 60.) ;
	seconds	= (gint) (temp
			- years * 365.25 * 24. * 3600.
			- days *24. * 3600.
			- hours * 3600.
			- minutes * 60.) ;

	if ( years > 0 )
		return g_strdup_printf("%d years %d days %d hours %d minutes %d seconds", years, days, hours, minutes, seconds) ;
	else if ( days > 0 )
		return g_strdup_printf("%d days %d hours %d minutes %d seconds", days, hours, minutes, seconds) ;
	else if ( hours > 0 )
		return g_strdup_printf("%d hours %d minutes %d seconds", hours, minutes, seconds) ;
	else /* Hey, you're computing a workunit in less than an hour! Great!*/
		return g_strdup_printf("%d minutes %d seconds", minutes, seconds) ;
}

/** Try to collect information about the user.
 * 
 * Information about the user will be displayed in the
 * config tab of the plugin.
 */
void get_user_info(void)
{
	FILE	*f ;
	gchar	*index1, *index2 ;
	gint	 length ;


	if ( ! seti_paths_status ) return ;
	
	if ( user_info.name != NULL ) {
#ifdef DEBUG
		if ( GK.debug ) printf("SETI@home plugin: user info already filled in\n") ;
#endif
		return ;
	}
	
	if ( (f = fopen(client_info.user_file, "r")) != NULL ) {
		while ( (fgets(buf, sizeof(buf), f)) != NULL ) {
			if ( strncmp(buf, "email_addr=", 11) == 0 ) {
				user_info.email_addr = g_strdup(buf+11) ;
			} else if ( strncmp(buf, "name=", 5) == 0 ) {
				user_info.name = g_strdup(buf+5) ;
			} else if ( strncmp(buf, "url=", 4) == 0 ) {
				user_info.url = g_strdup(buf+4) ;
			} else if ( strncmp(buf, "country=", 8) == 0 ) {
				user_info.country = g_strdup(buf+8) ;
			} else if ( strncmp(buf, "postal_code=", 12) == 0 ) {
				user_info.postal_code = g_strdup(buf+12) ;
			} else if ( strncmp(buf, "register_time=", 14) == 0 ) {
				index1 = strchr(buf, '(') ;
				index2 = strrchr(buf, ')') ;
				if ( NULL == index1 || NULL == index2 ) user_info.register_time = g_strdup(buf+14) ;
				else {
					length = index2 - index1 ;
					index1 ++ ;
					user_info.register_time = g_strndup(index1, length) ;
					user_info.register_time[length-1] = '\0' ;
				}
			/*
			} else if ( strncmp(buf, "last_wu_time=", 13) == 0 ) {
				user_info.last_wu_time = g_strdup(buf+13) ;
			*/
			} else if ( strncmp(buf, "last_result_time=", 17) == 0 ) {
				index1 = strchr(buf, '(') ;
				index2 = strrchr(buf, ')') ;
				if ( NULL == index1 || NULL == index2 ) user_info.last_result_time = g_strdup(buf+14) ;
				else {
					length = index2 - index1 ;
					index1 ++ ;
					user_info.last_result_time = g_strndup(index1, length) ;
					user_info.last_result_time[length-1] = '\0' ;
				}
			} else if ( strncmp(buf, "total_cpu=", 10) == 0 ) {
				/* user_info.total_cpu = g_strdup(buf+10) ; */
				user_info.total_cpu = strtotime(buf+10) ;
			} else if ( strncmp(buf, "nresults=", 9) == 0 ) {
				user_info.nresults = g_strdup(buf+9) ;
			}
		}
		fclose(f) ;
	} else {
		user_info.nresults = 0 ;
	}
}

/* To avoid memory leaks, the user_info structure
 * is freed whenever needed. All the actions
 * required to do this are gathered in this
 * function.
 */
void clean_user_info()
{
	g_free(user_info.email_addr) ;
	user_info.email_addr = NULL ;
	
	g_free(user_info.name) ;
	user_info.name = NULL ;
	
	g_free(user_info.url) ;
	user_info.url = NULL ;
	
	g_free(user_info.country) ;
	user_info.country = NULL ;
	
	g_free(user_info.postal_code) ;
	user_info.postal_code = NULL ;

	g_free(user_info.register_time) ;
	user_info.register_time = NULL ;

	g_free(user_info.last_wu_time) ;
	user_info.last_wu_time = NULL ;

	g_free(user_info.last_result_time) ;
	user_info.last_result_time = NULL ;

	g_free(user_info.total_cpu) ;
	user_info.total_cpu = NULL ;

	g_free(user_info.nresults) ;
	user_info.nresults = NULL ;
}

/**
 * returns the work unit status:
 *  0 = no work unit found
 *  1 = beeing downloaded
 *  2 = beeing processed
 *  3 = completed, waiting for the internet to come ;)
 */
void get_work_unit_status()
{
	FILE		*f ;
	gchar		*work_unit ;
	gint		 prog_found ;


	if ( ! seti_paths_status ) return ;
	
	work_unit = g_strdup_printf("%s/wtemp.sah", client_info.seti_file_path) ;
	if ( (f = fopen(work_unit, "r")) != NULL ) {
		g_free(work_unit) ;
		if ( fgets(buf, sizeof(buf), f) != NULL ) {
			if ( strncmp(buf, "type=work unit", 14) == 0 ) {
				fclose(f) ;
				work_unit_info.lprogress = 0. ;
				work_unit_info.status = 1 ;
				return ;
			} else if ( strncmp(buf, "type=result", 11) == 0 ) {
				fclose(f) ;
				work_unit_info.lprogress = 100. ;
				work_unit_info.status = 3 ;
				return ;
			}
		}
	}
	g_free(work_unit) ;
	
	prog_found = 0 ;
	if ( (f = fopen(client_info.state_file, "r")) != NULL ) {
		while ( (fgets(buf, sizeof(buf), f)) != NULL ) {
			if ( strncmp(buf, "prog", 4) == 0 ) {
				prog_found = 1 ;

				if ( locale_conv && locale_conv->decimal_point ) buf[6] = locale_conv->decimal_point[0] ;
				sscanf(buf,"prog=%lg", &(work_unit_info.lprogress));

				work_unit_info.lprogress *= 100. ;
			}
		}
		fclose(f);

		if ( ! prog_found ) work_unit_info.lprogress = 0. ;

		work_unit_info.status = 2 ;
		return ;
	}

	work_unit = g_strdup_printf("%s/work_unit.sah", client_info.seti_file_path) ;
	if ( (f = fopen(work_unit, "r")) != NULL ) {
		g_free(work_unit) ;
		if ( fgets(buf, sizeof(buf), f) != NULL ) {
			if ( strncmp(buf, "type=work unit", 14) == 0 ) {
				fclose(f) ;
				work_unit_info.lprogress = 0. ;
				work_unit_info.status = 2 ;
				return ;
			}
		}
	}
	g_free(work_unit) ;

	work_unit_info.status = 0 ;
	return ;
}

/** Gather information about the process.
 *
 *
 */
gdouble get_pcpu_info(gint pid)
{
	FILE *f;
	static gint	counter;
	static gchar	token[2][20];
	gchar		*tokenX;
	static gdouble	utime;
	static gdouble	stime;
	static gdouble	starttime;
	static gdouble	uptime;
	static gdouble	pcpu;
	
	
	if ((f = fopen("/proc/uptime", "r")) != NULL) {
		fscanf(f, "%lf", &uptime);
		fclose(f);
	}
	else return -1;
	
	if ((f = fopen(g_strdup_printf("/proc/%d/stat", pid), "r")) != NULL) {
		fgets(buf, sizeof(buf), f);
		fclose(f);
		
		tokenX = strtok(buf," ");
		for (counter = 0; counter < 22; counter++) {
			if (tokenX != NULL) {
				switch (counter) {
					case 13: strcpy(token[0], tokenX);
					break;
					case 14: strcpy(token[1], tokenX);
					break;
					case 21: strcpy(token[2], tokenX);
					break;
				}
				tokenX = strtok(NULL, " ");
			}
		}
	}
	else return -1;
	
	utime = atol(token[0]);
	stime = atol(token[1]);
	starttime = atol(token[2]) / HERTZ;
	
	total_time[my_switch] = (utime + stime) / HERTZ;
	seconds[my_switch] = uptime - starttime;
	
	
	if (my_switch) {
		total_time[2] = total_time[1] - total_time[0];
		seconds[2] = seconds[1] - seconds[0];
		
		/*
		 * this fails, but why? :)
		pcpu = ((total_time[1] - total_time[0]) * 100) / (seconds[1] - seconds[0]);
		 */
		pcpu = (total_time[2] * 100) / seconds[2];
		
		if (pcpu > 99.9) pcpu = 99.9;
	}
	
	my_switch = (my_switch == 0) ;

	return pcpu;
}



